const fs = require("fs");
const path = require("path");
const chalk = require("chalk");
const globby = require("globby");
const { cli } = require("cli-ux");
const inquirer = require("inquirer");
const Shopify = require("shopify-api-node");
const { Command } = require("@oclif/command");
const { Config } = require("@dotdev/next-config");
const { Deployer } = require("@dotdev/next-deployer");

class ConnectCommand extends Command {
  static shopify(config) {
    const environment = config.getEnvironment();

    return new Shopify({
      shopName: environment.domains.shopify,
      apiKey: environment.credentials.apiKey,
      password: environment.credentials.password
    });
  }

  async run() {
    this.log("Welcome to Next!");
    this.log();

    const config = Config.resolve();

    await this.environment(config);

    await this.credentials(config);

    await this.theme(config);

    await this.assetBase(config);

    await this.cert(config);

    this.log(chalk.yellow("You're all set to go! 🎉"));
    this.log();
    this.log(
      chalk.gray(
        `You may wish to run ${chalk.white(
          config.getScript("fresh")
        )} if your theme is stale;`
      )
    );
    this.log(
      chalk.gray(
        `or ${chalk.white(
          config.getScript("dev")
        )} to begin building awesome things!`
      )
    );
  }

  async environment(config) {
    const environments = config.get("environments");

    if (!environments) {
      this.error(
        "No environments defined, check your project next.config.json file"
      );
    }

    const handles = Object.keys(environments);

    if (1 == handles.length) {
      await config.setEnvironment(handles.shift());

      const environment = config.getEnvironment();

      this.log(
        `Auto selected environment: ${chalk.green(
          environment.handle + " (" + environment.domains.shopify + ")"
        )}`
      );

      this.log();

      return;
    }

    const choices = handles.map(handle => ({
      name: `${handle} (${environments[handle].domains.shopify})`,
      value: handle
    }));

    if (config.hasEnvironment()) {
      const { handle } = config.getEnvironment();
      choices.sort((x, y) =>
        handle == x.value ? -1 : handle == y.value ? 1 : 0
      );
    }

    const answers = await inquirer.prompt({
      type: "list",
      name: "environment",
      message: "Select an environment to connect",
      choices,
      default: config.hasEnvironment() ? config.getEnvironment().handle : null
    });

    this.log();

    await config.setEnvironment(answers.environment);
  }

  async credentials(config) {
    const environment = config.getEnvironment();

    const answers = await inquirer.prompt([
      {
        name: "apiKey",
        message: "Enter your API key",
        default: environment.credentials ? environment.credentials.apiKey : null
      },
      {
        name: "password",
        message: "Enter your API password",
        default: environment.credentials
          ? environment.credentials.password
          : null
      }
    ]);

    this.log();

    await config
      .set(`environments.${environment.handle}.credentials`, answers)
      .save();
  }

  async theme(config) {
    const shopify = ConnectCommand.shopify(config);

    cli.action.start("Fetching themes from Shopify");

    const themes = await shopify.theme.list();

    cli.action.stop();

    this.log();

    const choices = themes.map(theme => ({
      name: `${theme.name} (${theme.id})${
        "main" === theme.role ? " " + chalk.bgRed("Published") : ""
      }`,
      value: theme
    }));

    if (config.hasTheme()) {
      const { id } = config.getTheme();
      choices.sort((x, y) =>
        id == x.value.id ? -1 : id == y.value.id ? 1 : 0
      );
    }

    const answers = await inquirer.prompt({
      type: "list",
      name: "theme",
      message: "Select a theme to connect",
      choices
    });

    this.log();

    await config.setTheme(answers.theme);
  }

  async assetBase(config) {
    const assetBaseKey = [
      "environments",
      config.getEnvironment().handle,
      "themes",
      config.getTheme().id,
      "assetBase"
    ].join(".");

    let assetBase = config.get(assetBaseKey);

    if (!assetBase) {
      const deployer = new Deployer(config);

      cli.action.start("Fetching CDN asset base for theme");

      assetBase = await deployer.getAssetBase(config.getTheme().id);

      cli.action.stop();

      this.log();

      config.set(assetBaseKey, assetBase).save();
    }

    this.log(`Asset base: ${chalk.green(assetBase)}`);
    this.log();
  }

  async cert(config) {
    const existingCertificate = config.getCertificate();

    if (
      existingCertificate &&
      fs.existsSync(existingCertificate.cert) &&
      fs.existsSync(existingCertificate.key)
    ) {
      this.log(`Certificate: ${chalk.green(existingCertificate.cert)}`);
      this.log(`Key: ${chalk.green(existingCertificate.key)}`);
      this.log();
      return;
    }

    cli.action.start("Scanning for certificate");

    const certificate = await this.scanForCert(process.cwd());

    cli.action.stop();

    this.log();

    if (certificate) {
      await config.setCertificate(certificate);
      this.log(`Certificate: ${chalk.green(certificate.cert)}`);
      this.log(`Key: ${chalk.green(certificate.key)}`);
      this.log();
      return;
    }

    this.log(chalk.red(`Certificate could not be found in ${process.cwd()}`));

    this.log();

    this.log(
      "To run your theme locally, you will need to generate a trusted certificate for:"
    );
    this.log(`${chalk.blue(config.getEnvironment().domains.local)}`);

    this.log();

    this.log(
      `We recommend installing "mkcert" and running the following within your project directory:`
    );

    this.log(chalk.blue("mkcert " + config.getEnvironment().domains.local));

    this.log();

    this.log(
      `Download: ${chalk.blue("https://github.com/FiloSottile/mkcert")}`
    );

    this.log();

    const answers = await inquirer.prompt({
      name: "again",
      type: "confirm",
      message: "Do you want to scan again for a certificate?"
    });

    this.log();

    if (answers.again) {
      return await this.cert(config);
    }

    this.error("Cannot run locally without trusted certificate");
  }

  async scanForCert(dir) {
    const paths = await globby("*.pem");

    const cert = { cert: null, key: null };

    paths.forEach(file => {
      const data = fs.readFileSync(file).toString();

      let component;

      if (data.includes("BEGIN CERTIFICATE")) component = "cert";
      else if (data.includes("BEGIN PRIVATE KEY")) component = "key";
      else return;

      cert[component] = path.resolve(process.cwd(), file);
    });

    if (!cert.cert || !cert.key) {
      return null;
    }

    return cert;
  }
}

ConnectCommand.description = "connect to an environment";

module.exports = ConnectCommand;
